/*
 * include/asm-powerpc/Bestcomm.h
 *
 * Driver for MPC52xx processor BestComm peripheral controller
 * Using bplan GmbH OpenFirmware
 *
 * 2006 (c) bplan GmbH  This file is licensed under
 * the terms of the GNU General Public License version 2.  This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 *
 */

#ifndef __BESTCOMM_BESTCOMM_H__
#define __BESTCOMM_BESTCOMM_H__


/**************/
/**************/
/**************/

#include <asm/io.h>
#include <asm/mpc52xx.h>

#include <linux/types.h>

/**************/
/**************/
/**************/

#define BESTCOMM_MAX_VAR    24
#define BESTCOMM_MAX_INC    8
#define BESTCOMM_MAX_TASKS  16

/**************/
/**************/
/**************/

/* Task Descriptor Table Entry */
/* copied from bestcomm.h, can be found in the public freescale doc */

/* pragma pack required ? */
struct bestcomm_tdt {
	u32 start;
	u32 stop;
	u32 var;
	u32 fdt;
	u32 exec_status; /* used internally by SmartComm engine */
	u32 mvtp;		 /* used internally by SmartComm engine */
	u32 context;
	u32 litbase;
};

struct bestcomm_taskhandle
{
	int taskid;
	int irq;

	struct bestcomm_tdt __iomem * bestcomm_tdt;

	u32 __iomem *bestcomm_taskcode;
	u32 __iomem *bestcomm_vartable;

	union {
		struct sdma_bd *bd;
		struct sdma_bd2 *bd2;
	};

	void **cookie;
	u16 index;
	u16 outdex;
	u16 num_bd;
	u32 flags;
};


/**************/
struct bestcomm_mainhandle
{
	struct device_node *buildin_ofwnode;
	struct device_node *sram_ofwnode;
	struct device_node *bestcomm_ofwnode;

	int bestcomm_irq;

	unsigned long sdma_io_basebus;
	struct mpc52xx_sdma __iomem* sdma_io_basevirt;
	size_t sdma_io_size;

	struct bestcomm_tdt __iomem *sdma_tdtentry;

	unsigned long sram_basebus;
	void __iomem* sram_basevirt;
	size_t sram_size;
	size_t sram_available;

	struct bestcomm_taskhandle taskhandle_table[BESTCOMM_MAX_TASKS];
};



/**************/
/**************/
/**************/
int
bestcomm_init_once(void);

struct device_node *
bestcomm_find_hw(char *name);

int 
bestcomm_hwiscapable(void);

u32
bestcomm_getreg(char *name, int *size);

int
bestcomm_getintrvector(char *name);

/**************/
struct bestcomm_taskhandle *
bestcomm_taskallocate(int bestcomm_tasknum, int queue_size);

void
bestcomm_taskfree(struct bestcomm_taskhandle *mytaskhandle);

int
bestcomm_taskclear_irq(struct bestcomm_taskhandle *mytaskhandle);

int
bestcomm_taskenable(struct bestcomm_taskhandle *mytaskhandle);

int
bestcomm_taskdisable(struct bestcomm_taskhandle *mytaskhandle);

int
bestcomm_taskget_irq(struct bestcomm_taskhandle *mytaskhandle);

u32 __iomem*
bestcomm_taskget_code(struct bestcomm_taskhandle *mytaskhandle);

u32 __iomem*
bestcomm_taskget_vartable(struct bestcomm_taskhandle *mytaskhandle);

u32 __iomem*
bestcomm_taskget_inctable(struct bestcomm_taskhandle *mytaskhandle);

u16 __iomem*
bestcomm_taskget_tcr(struct bestcomm_taskhandle *mytaskhandle);

void __iomem *
bestcomm_phys_to_virt(struct bestcomm_mainhandle *mymainhandle, unsigned long busaddr);

unsigned long
bestcomm_virt_to_phys(struct bestcomm_mainhandle *mymainhandle, void *virtaddr);

void __iomem*
bestcomm_sram_alloc(struct bestcomm_taskhandle *mytaskhandle, int len, unsigned long *busaddr);

void
bestcomm_sram_free(struct bestcomm_taskhandle *mytaskhandle, void *ptr, int len);

/**************/
/*
 * The original Linux API has been here partially copied or wrapped
 * this way, it's quiet easy to reuse exising code
*/
/* Buffer Descriptor definitions */
struct sdma_bd {
	u32 status;
	void *data;
};

struct sdma_bd2 {
	u32 status;
	void *data1;
	void *data2;
};

#define FIELD_OFFSET(s,f) ((unsigned long)(&(((struct s*)0)->f)))

#define SDMA_FLAGS_NONE		0x0000
#define SDMA_FLAGS_ENABLE_TASK	0x0001
#define SDMA_FLAGS_BD2		0x0002
#define SDMA_BD_READY		0x40000000UL

#define SDMA_FEC_TX_BD_TFD	0x08000000UL	/* transmit frame done */
#define SDMA_FEC_TX_BD_INT	0x04000000UL	/* Interrupt */
#define SDMA_FEC_TX_BD_TFD_INIT	(SDMA_BD_READY | SDMA_FEC_TX_BD_TFD | SDMA_FEC_TX_BD_INT)

#define SDMA_LEN_BITS		26
#define SDMA_LEN_MASK		((1 << SDMA_LEN_BITS) - 1)

#define SDMA_BD_ALIGN		0x10

/**************/
/**************/
/**************/
struct sdma { struct bestcomm_taskhandle taskhandle; };

/**************/
/**************/
/**************/
unsigned long
sdma_sram_pa(void __iomem *virt);

void __iomem *
sdma_sram_va(unsigned long pa);

unsigned long sdma_io_pa
(void __iomem *virt);

void __iomem *
sdma_io_va(unsigned long pa);

/**************/
struct sdma *
sdma_fex_tx_preinit(int bdnum);

struct sdma *
sdma_fex_rx_preinit(int bdnum);

extern int
sdma_fec_rx_init(struct sdma *s, phys_addr_t fifo, int maxbufsize);

extern int
sdma_fec_tx_init(struct sdma *s, phys_addr_t fifo);

struct sdma *
sdma_ata_preinit(int maxbuffers);

int
sdma_ata_init(struct bestcomm_taskhandle *mytaskhandle, int maxbufsize);

struct sdma_ata_var {
	u32 enable;		/* (u16*) address of task's control register */
	u32 bd_base;		/* (struct sdma_bd*) beginning of ring buffer */
	u32 bd_last;		/* (struct sdma_bd*) end of ring buffer */
	u32 bd_start;		/* (struct sdma_bd*) current bd */
	u32 buffer_size;	/* size of receive buffer */
};

/**************/
/* ata task incs that need to be set before enabling the task */
struct sdma_ata_inc {
	u16 pad0;
	s16 incr_bytes;
	u16 pad1;
	s16 incr_dst;
	u16 pad2;
	s16 incr_src;
};

/* rx task vars that need to be set before enabling the task */
struct sdma_fec_rx_var {
	u32 enable;		/* (u16*) address of task's control register */
	u32 fifo;		/* (u32*) address of fec's fifo */
	u32 bd_base;		/* (struct sdma_bd*) beginning of ring buffer */
	u32 bd_last;		/* (struct sdma_bd*) end of ring buffer */
	u32 bd_start;		/* (struct sdma_bd*) current bd */
	u32 buffer_size;	/* size of receive buffer */
};

/* rx task incs that need to be set before enabling the task */
struct sdma_fec_rx_inc {
	u16 pad0;
	s16 incr_bytes;
	u16 pad1;
	s16 incr_dst;
	u16 pad2;
	s16 incr_dst_ma;
};

/* tx task vars that need to be set before enabling the task */
struct sdma_fec_tx_var {
	u32 DRD;		/* (u32*) address of self-modified DRD */
	u32 fifo;		/* (u32*) address of fec's fifo */
	u32 enable;		/* (u16*) address of task's control register */
	u32 bd_base;		/* (struct sdma_bd*) beginning of ring buffer */
	u32 bd_last;		/* (struct sdma_bd*) end of ring buffer */
	u32 bd_start;		/* (struct sdma_bd*) current bd */
	u32 buffer_size;	/* set by uCode for each packet */
};

/* tx task incs that need to be set before enabling the task */
struct sdma_fec_tx_inc {
	u16 pad0;
	s16 incr_bytes;
	u16 pad1;
	s16 incr_src;
	u16 pad2;
	s16 incr_src_ma;
};
/**************/
void *
sdma_sram_alloc(int size, int alignment, u32 *dma_handle);

static inline struct sdma *
sdma_alloc(int request_queue_size)
{ return NULL; };

static inline void
sdma_free(struct sdma *s)
{  bestcomm_taskfree( (struct bestcomm_taskhandle *)s); }

static inline int
sdma_irq(struct sdma *s)
{ return bestcomm_taskget_irq(&s->taskhandle); }

static inline void
sdma_enable(struct sdma *s)
{ bestcomm_taskenable(&s->taskhandle); }

static inline void
sdma_disable(struct sdma *s)
{ bestcomm_taskdisable(&s->taskhandle); }

static inline int
sdma_queue_empty(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;
	return s->index == s->outdex;
}

static inline void
sdma_clear_irq(struct sdma *s)
{ bestcomm_taskclear_irq( &s->taskhandle); }

static inline int
sdma_next_index(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;
	return ((s->index + 1) == s->num_bd) ? 0 : s->index + 1;
}

static inline int
sdma_next_outdex(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;
	return ((s->outdex + 1) == s->num_bd) ? 0 : s->outdex + 1;
}

static inline int
sdma_queue_full(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;
	return s->outdex == sdma_next_index(s_);
}

static inline int
sdma_buffer_done(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;

	if (sdma_queue_empty(s_))
		return 0;

	rmb();
	return (s->bd[s->outdex].status & SDMA_BD_READY) == 0;
}

static inline int
sdma_buffer_done_fixed(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;
	rmb();
	return (s->bd[s->outdex].status & SDMA_BD_READY) == 0;
}

static inline int
sdma_buffer2_done(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;

	if (sdma_queue_empty(s_))
		return 0;

	rmb();
	return (s->bd2[s->outdex].status & SDMA_BD_READY) == 0;
}

static inline void
sdma_submit_buffer(struct sdma *s_, void *cookie, void *data, int length)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;

	s->cookie[s->index] = cookie;
	s->bd[s->index].data = data;
	wmb();
	s->bd[s->index].status = SDMA_BD_READY | length;
	s->index = sdma_next_index(s_);
	if (s->flags & SDMA_FLAGS_ENABLE_TASK)
		bestcomm_taskenable(s);
}

/*
 * Special submit_buffer function to submit last buffer of a frame to
 * the FEC tx task.  tfd means "transmit frame done".
 */
static inline void
sdma_fec_tfd_submit_buffer(struct sdma *s_, void *cookie, void *data, int length)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;

	s->cookie[s->index] = cookie;
	s->bd[s->index].data = data;
	wmb();
	s->bd[s->index].status = SDMA_FEC_TX_BD_TFD_INIT | length;
	s->index = sdma_next_index(s_);
	bestcomm_taskenable(s);
}

static inline void *
sdma_retrieve_buffer(struct sdma *s_, int *length)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;
	void *cookie = s->cookie[s->outdex];

	if (length) {
		rmb();
		*length = s->bd[s->outdex].status & SDMA_LEN_MASK;
	}
	s->outdex = sdma_next_outdex(s_);
	return cookie;
}

static inline void
sdma_reset_buffers(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;

	while (!sdma_queue_empty(s_)) {
		s->bd[s->outdex].status = 0;
		wmb();
		s->bd[s->outdex].data = 0;
		sdma_retrieve_buffer(s_, NULL);
	}
	s->index = s->outdex = 0;
}

static inline void
sdma_submit_buffer2(struct sdma *s_, void *cookie,
					void *data1, void *data2, int length)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;

	s->cookie[s->index] = cookie;
	s->bd2[s->index].data1 = data1;
	s->bd2[s->index].data2 = data2;
	wmb();
	s->bd2[s->index].status = SDMA_BD_READY | length;
	s->index = sdma_next_index(s_);
	if (s->flags & SDMA_FLAGS_ENABLE_TASK)
		bestcomm_taskenable(s);
}

static inline void *
sdma_retrieve_buffer2(struct sdma *s_, int *length)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;
	void *cookie = s->cookie[s->outdex];

	if (length) {
		rmb();
		*length = s->bd2[s->outdex].status & SDMA_LEN_MASK;
	}

	s->outdex = sdma_next_outdex(s_);
	return cookie;
}

static inline void
sdma_reset_buffers2(struct sdma *s_)
{
	struct bestcomm_taskhandle *s = &s_->taskhandle;
	while (!sdma_queue_empty(s_)) {
		s->bd2[s->outdex].status = 0;
		s->bd2[s->outdex].data1 = 0;
		s->bd2[s->outdex].data2 = 0;
		wmb();
		sdma_retrieve_buffer2(s_, NULL);
	}

	s->index = s->outdex = 0;
}

#endif 

