/*
 * arch/powerpc/platforms/mpc52xx_bestcomm_helper.c
 *
 * This piece of code help the driver using the Bestcomm DMA
 * for example, it find and setup the bestcomm dma tasks.
 * Notice that the prototypes are all gather in Bestcomm.h
 *
 * 2006 (c) bplan GmbH  This file is licensed under
 * the terms of the GNU General Public License version 2.  This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 *
 */

#undef DEBUG

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/module.h>

#include <asm/mpc52xx.h>
#include <asm/prom.h>
#include <asm/irq.h>

#include "bestcomm.h"


/**************/
/**************/
/**************/
extern struct bestcomm_mainhandle bestcomm_mainhandle;

static int irqhack(int irq)
{
        struct irq_host *mpc52xx_irqhost;
        struct device_node *pic_dev;

        pic_dev = of_find_compatible_node(NULL, "interrupt-controller", "mpc5200-pic");

        if (pic_dev == NULL)
                return -1;

        mpc52xx_irqhost = irq_find_host(pic_dev);
        if (mpc52xx_irqhost == NULL)
                return -1;

        pr_debug("%s: irq=0x%x\n", __func__, irq);
        return irq_create_mapping(mpc52xx_irqhost, irq);
}


/**************/
/**************/
/**************/
static struct sdma *
sdma_task_preinit(char * devicetype, char * devicecomp, char* taskname, int bdnum)
{
        struct sdma *sdma;
        struct device_node *onenode;
	struct device_node *parentnode;
        uint32_t *oneptr;
        int tasknum;
	int irq;

        pr_debug("%s:\n", __FUNCTION__);

        if (!devicecomp)
                return NULL;

        if (!taskname)
                return NULL;

        if ( bestcomm_init_once() != 0)
                return NULL;

        pr_debug("devicecomp=%s, taskname=%s, type=%s\n", devicecomp, taskname, devicetype);

	onenode = of_find_compatible_node(NULL, devicetype, devicecomp);
        if (!onenode)
                return NULL;

        pr_debug("node=%p\n", onenode);

	parentnode = onenode;
	onenode = NULL;
	while( (onenode = of_get_next_child(parentnode, onenode) ) )
	{
	        pr_debug("node=%p, name=%s\n", onenode, onenode->name);
		if ( strcmp(onenode->name, taskname) == 0)
			break;
	}

        pr_debug("node=%p\n", onenode);

	if (!onenode)
                return NULL;

        oneptr = (uint32_t*) get_property(onenode, "taskid", NULL);
        if (!oneptr)
                return NULL;

        tasknum = (int) * oneptr;

        pr_debug("tasknum=%d, bdnum=%d\n", tasknum, bdnum);

        sdma = (struct sdma *) bestcomm_taskallocate(tasknum, bdnum);
        if (!sdma)
                return NULL;

	irq = irq_of_parse_and_map(onenode, 0);
	irqhack(irq);

        return sdma;
}


/**************/
/**************/
/**************/
/*
 * FEC driver helper
*/

struct sdma *
sdma_fex_tx_preinit(int bdnum)
{
        pr_debug("\n");
        return sdma_task_preinit("network", "mpc5200-ethernet", "bestcomm-txtask", bdnum);
}

int
sdma_fec_tx_init(struct sdma *s, phys_addr_t fifo)
{
        struct bestcomm_taskhandle *mytaskhandle = (struct bestcomm_taskhandle *) s;
        struct sdma_fec_tx_var *var;
        struct sdma_fec_tx_inc *inc;

        int tasknum = -1;
        struct sdma_bd *bd = 0;
        u32 bd_pa;

        if (!bd)
                bd = (struct sdma_bd *)sdma_sram_alloc(sizeof(*bd) * mytaskhandle->num_bd, SDMA_BD_ALIGN, &bd_pa);
        if (!bd)
                return -ENOMEM;

        bestcomm_taskdisable(mytaskhandle);

        tasknum = mytaskhandle->taskid;

        mytaskhandle->bd = bd;
        mytaskhandle->flags = SDMA_FLAGS_ENABLE_TASK;
        mytaskhandle->index = 0;
        mytaskhandle->outdex = 0;
        memset(bd, 0, sizeof(*bd) * mytaskhandle->num_bd);

        var = (struct sdma_fec_tx_var *) bestcomm_taskget_vartable(mytaskhandle);
        var->DRD	= sdma_sram_pa( bestcomm_taskget_code(mytaskhandle) + 31);
        var->fifo	= fifo;
        var->enable	= sdma_io_pa(&bestcomm_mainhandle.sdma_io_basevirt->tcr[tasknum]);
        var->bd_base	= bd_pa;
        var->bd_last	= bd_pa + (mytaskhandle->num_bd - 1)*sizeof(struct sdma_bd);
        var->bd_start	= bd_pa;

        /* These are constants, they should have been in the image file */
        inc = (struct sdma_fec_tx_inc *)bestcomm_taskget_inctable(mytaskhandle);
        inc->incr_bytes	= -(s16)sizeof(u32);
        inc->incr_src		= sizeof(u32);
        inc->incr_src_ma	= sizeof(u8);

        pr_debug("tasknum=%d, bdnum=%d\n", tasknum, mytaskhandle->num_bd);

        mb();

        return tasknum;
}

struct sdma *sdma_fex_rx_preinit(int bdnum)
{
        pr_debug("%s:\n", __FUNCTION__);
        return sdma_task_preinit("network", "mpc5200-ethernet", "bestcomm-rxtask", bdnum);
}


int sdma_fec_rx_init(struct sdma *s, phys_addr_t fifo, int maxbufsize)
{
        int tasknum;
        struct sdma_fec_rx_var *var;
        struct sdma_fec_rx_inc *inc;
        struct bestcomm_taskhandle *mytaskhandle = (struct bestcomm_taskhandle *) s;

        struct sdma_bd *bd = 0;
        u32 bd_pa;

        if (!bd)
                bd = (struct sdma_bd *)sdma_sram_alloc(sizeof(*bd) * mytaskhandle->num_bd, SDMA_BD_ALIGN, &bd_pa);

        if (!bd)
                return -ENOMEM;

        bestcomm_taskdisable(mytaskhandle);

        tasknum = mytaskhandle->taskid;

        mytaskhandle->bd = bd;
        mytaskhandle->flags = SDMA_FLAGS_NONE;
        mytaskhandle->index = 0;
        mytaskhandle->outdex = 0;
        memset(bd, 0, sizeof(*bd) * mytaskhandle->num_bd);

        var = (struct sdma_fec_rx_var *) bestcomm_taskget_vartable(mytaskhandle);

        var->enable		= sdma_io_pa(&bestcomm_mainhandle.sdma_io_basevirt->tcr[tasknum]);
        var->fifo		= fifo;
        var->bd_base		= bd_pa;
        var->bd_last		= bd_pa + (mytaskhandle->num_bd - 1)*sizeof(struct sdma_bd);
        var->bd_start		= bd_pa;
        var->buffer_size	= maxbufsize;

        /* These are constants, they should have been in the image file */
        inc = (struct sdma_fec_rx_inc *) bestcomm_taskget_inctable(mytaskhandle);
        inc->incr_bytes	= -(s16)sizeof(u32);
        inc->incr_dst		= sizeof(u32);
        inc->incr_dst_ma	= sizeof(u8);

        pr_debug("tasknum=%d, bdnum=%d\n", tasknum, mytaskhandle->num_bd);

        mb();

        return tasknum;
}

/**************/
/**************/
/**************/
int sdma_ata_init(struct bestcomm_taskhandle *mytaskhandle, int maxbufsize)
{
        struct sdma_ata_var *var;
        struct sdma_bd2 *bd2 = 0;
        u32 bd_pa;
        int tasknum = -1;

        pr_debug("MyTaskHandle=%pn max buf=%d\n", mytaskhandle, maxbufsize);

        if (!bd2)
                bd2 = (struct sdma_bd2 *)sdma_sram_alloc(sizeof(*bd2) * mytaskhandle->num_bd, SDMA_BD_ALIGN, &bd_pa);

        if (!bd2)
                return -ENOMEM;

        bestcomm_taskdisable(mytaskhandle);

        tasknum = mytaskhandle->taskid;

        mytaskhandle->flags = SDMA_FLAGS_BD2;
        mytaskhandle->bd2 = bd2;
        mytaskhandle->index = 0;
        mytaskhandle->outdex = 0;
        memset(bd2, 0, sizeof(*bd2) * mytaskhandle->num_bd);

        var = (struct sdma_ata_var *)bestcomm_taskget_vartable(mytaskhandle);
        var->enable		= sdma_io_pa(&bestcomm_mainhandle.sdma_io_basevirt->tcr[tasknum]);
        var->bd_base		= bd_pa;
        var->bd_last		= bd_pa + (mytaskhandle->num_bd - 1)*sizeof(struct sdma_bd2);
        var->bd_start		= bd_pa;
        var->buffer_size	= maxbufsize;

        mb();

        return 0;
}

/**************/

struct sdma *sdma_ata_preinit(int maxbuffers)
{
        pr_debug("%s:\n", __FUNCTION__);
        return sdma_task_preinit("ata", "mpc5200-ata", "bestcomm-task", maxbuffers);
}


/**************/
/*
 * Export symbol for modules
*/

EXPORT_SYMBOL(sdma_sram_pa);
EXPORT_SYMBOL(sdma_sram_va);
EXPORT_SYMBOL(sdma_io_pa);
EXPORT_SYMBOL(sdma_io_va);

EXPORT_SYMBOL(sdma_fex_tx_preinit);
EXPORT_SYMBOL(sdma_fex_rx_preinit);
EXPORT_SYMBOL(sdma_fec_rx_init);
EXPORT_SYMBOL(sdma_fec_tx_init);

EXPORT_SYMBOL(sdma_ata_preinit);
EXPORT_SYMBOL(sdma_ata_init);
