/*
 *  mpc52xx_atablock / ata.c
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */


/**************/
/**************/
/**************/

#include "mpc52xx_blockata.h"

/**************/
/**************/
/**************/
int mpc52xx_ata_dodrivereset(struct mpc52xx_blockata_priv  *priv)
{
        int ret;

        write_cmd(priv, ATA_CMD_RESET);

        ret = wait_not_busy(priv);
        if (ret<0)
                return ret;

        if ( ATASTS_GOT_ERR(read_altstatus(priv)) )
                return -1;

        return mpc52xx_ata_doidentify(priv);
}


/**************/
int mpc52xx_ata_regcheck(struct mpc52xx_blockata_priv  *priv)
{
        u8 seccnt;
        u8 sector;

        write_sectorcnt(priv, 0x55);
        write_sectornum(priv, 0xaa);
        write_sectorcnt(priv, 0xaa);
        write_sectornum(priv, 0x55);
        write_sectorcnt(priv, 0x55);
        write_sectornum(priv, 0xaa);

        seccnt = read_sectorcnt(priv);
        sector = read_sectornum(priv);

        NPRINTK("%s: seccnt=%x, sector=%x\n", __func__, seccnt,  sector );

        if
        (
                ( (seccnt == 0x55) || (seccnt == 0x01) )
                && (sector == 0xaa)
        )
                return 0;

        return -1;
}

/**************/
/**************/
static int ata_docpupollread(
        struct mpc52xx_blockata_priv  *priv,
        void *buffer,
        int len)
{
        u16 *buffer16 = (u16 *) buffer;
        int local_len = len;

        while(local_len--)
                *buffer16++ = read_data(priv);

        return len - local_len;
}

/**************/
static int ata_docpupollwrite(
        struct mpc52xx_blockata_priv  *priv,
        void *buffer,
        int len)
{
        u16 *buffer16 = (u16 *) buffer;
        int local_len = len;

        while(local_len--)
                write_data(priv, *buffer16++);

        return len - local_len;

}


/**************/
int mpc52xx_ata_setupsector(
        struct mpc52xx_blockata_priv  *priv,
        sector_t sector,
        int sector_num,
        int is_write)
{
        u8 seccnt, secnum, cyl_lo, cyl_hi, devhead;
        u16 cyl16;

        if (priv->drive_canlba) {
                if ( sector & (0xfffff0000000LL) ) {
                        if (!priv->drive_canlba48)
                                return -1;

                        write_sectornum(priv, (sector >> 24) & 0xff);

                        write_cyllo(priv, (sector >> 32) & 0xff);

                        write_cylhi(priv, (sector >> 40) & 0xff);

                        write_sectorcnt(priv, sector_num & 0xff00);
                        sector_num &= 0xff;

                        sector &= 0xffffff;

                        if ( (priv->multi_available) && (sector_num > 1) )
                                priv->curio_atacmd = is_write ? ATA_CMD_WRITE_MULTI_EXT : ATA_CMD_READ_MULTI_EXT;
                        else
                                priv->curio_atacmd = is_write ? ATA_CMD_PIO_WRITE_EXT : ATA_CMD_PIO_READ_EXT;

                } else {                        
                        if ( (priv->multi_available) && (sector_num > 1) )
                                priv->curio_atacmd = is_write ? ATA_CMD_WRITE_MULTI : ATA_CMD_READ_MULTI;
                        else
                                priv->curio_atacmd = is_write ? ATA_CMD_PIO_WRITE : ATA_CMD_PIO_READ;
                }

                secnum = sector & 0xff;
                cyl16 = (sector >> 8) & 0xffff;

                devhead = (sector >> 24) & 0xf;
                devhead |= ATA_LBA;
        } else {
                unsigned long blkno = (unsigned long) sector;
                int sectorspertrack = priv->drive_chs_sectorspertrack;
                int cylinders = priv->drive_chs_cylinders;
                int heads = priv->drive_chs_heads;

                secnum = (blkno % sectorspertrack) + 1;
                blkno -= secnum - 1;

                blkno /= sectorspertrack;

                devhead = blkno / cylinders;

                cyl16 = blkno % cylinders;

                if (devhead > (heads-1) ) {
                        return -1;
                }
        }

        seccnt = (u8) sector_num;
        cyl_hi = (cyl16 >> 8) & 0xff;
        cyl_lo = cyl16 & 0xff;

        write_devhead(priv, devhead);
        write_sectornum(priv, secnum);
        write_cyllo(priv, cyl_lo);
        write_cylhi(priv, cyl_hi);
        write_sectorcnt(priv, seccnt);

        NPRINTK("lba=%d, sector=%lld, seccnt=%x sector=0x%x (%d), cyl_lo=%x, cyl_hi=%x, cyl16=0x%x, devhead=%x\n",
                priv->drive_canlba, sector, seccnt, secnum, secnum, cyl_lo, cyl_hi, cyl16, devhead);

        return 0;
}



int mpc52xx_ata_doreset(struct mpc52xx_blockata_priv  *priv)
{
        int ret = 0;

        ret = wait_not_busy(priv);
        if (ret!=0)
                return -1;

        NPRINTK("%s: set reset %x\n", __func__, read_altstatus(priv) );
       
        write_ctl(priv, ATA_SRST | ATA_NIEN);
        ata_sleep(priv);

        NPRINTK("%s: release reset, %x\n", __func__, read_altstatus(priv));

        write_ctl(priv, ATA_NIEN);


        ata_sleep(priv);

        ret = wait_not_busy(priv);
        if (ret<0)
                return ret;

        write_devhead(priv, 0x00);
        ata_sleep(priv);

        return wait_not_busy(priv);
}

/*
 * Export stuff
*/

EXPORT_SYMBOL(mpc52xx_ata_setupsector);
EXPORT_SYMBOL(mpc52xx_ata_doreset);
EXPORT_SYMBOL(mpc52xx_ata_regcheck);
