/*
 *  mpc52xx_atablock / dodrivecmd.c
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */

/*
 * handle the ioctl HDIO_DRIVE_CMD command
*/

/**************/
/**************/
/**************/
#include "mpc52xx_blockata.h"

/**************/
/**************/
/**************/
irqreturn_t ata_drivecmd_handler(struct mpc52xx_blockata_priv *priv, u8 ata_status)
{
	NPRINTK("%s: status=0x%2.2x\n", __func__, ata_status);

	priv->ata_handler = ata_void_handler;
	priv->io_inprogress = 0;
	wake_up_interruptible(&priv->my_waitqueue);

                 return IRQ_HANDLED;
}

         int mpc52xx_ata_dodrivecmd(
                 struct mpc52xx_blockata_priv *priv,
                 unsigned long *irqflags,
                 u8 *args)
         {
                 int ret;
                 u8 status;
                 u8 cmd;

                 NPRINTK("arg [0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x],\n", args[0], args[1], args[2], args[3]);

                 cmd = args[0];
                 if ( (cmd==ATA_CMD_ID_ATAPI) || (cmd==ATA_CMD_ID_ATA) )
                 {
                         int i;
                         u16 *iddest_u16;
                         u16 *drive_idendify;

                         drive_idendify = priv->drive_identify;
                         iddest_u16 = (u16*) (&args[4]);
                         ret = 0;

                         if (!priv->drive_identify_valid) {
                                 if (priv->io_inprogress) {
                                         VPRINTK("IO already in-progress, can not do more!\n");
                                         ret = -EBUSY;
                                         goto end;
                                 }

                                 priv->io_inprogress = 1;
                                 ret = mpc52xx_ata_doidentify(priv);
                                 priv->io_inprogress = 0;

                                 if (ret!=0)
                                         goto end;
                         }

                         for(i=0; i < 256; i++)
                                 iddest_u16[i] = cpu_to_le16(drive_idendify[i]);

                         goto end;
                 }

                 if (priv->io_inprogress)
                 {
                         VPRINTK("IO already in-progress, can not do more!\n");
                         ret = -EBUSY;
                         goto end;
                 }

                 priv->io_inprogress = 1;

                 // a drive cmd looks very simple
                 // simply copy the stuff in the  drive reg and wait for some interrupt

                 // Get preprepatre to receive an interrupt
                 priv->ata_handler = ata_drivecmd_handler;

                 write_devfeature(priv, args[2]);
                 write_sectorcnt(priv, args[3]);
                 write_sectornum(priv, args[1]);

                 write_cmd(priv, cmd);

                 /*
                  * I'm 100% (well not even 10%) happy and confortable with this IRQ/wait stuff
                  * People should really write an how to "How to wait for an event in atomic section?"
                 */
                 spin_unlock_irqrestore(&priv->lock, *irqflags);
                 ret = wait_event_interruptible_timeout(priv->my_waitqueue, priv->io_inprogress==0, 5*HZ);
                 spin_lock_irqsave(&priv->lock, *irqflags);

                 if (ret<0)
                 {
                         ret = -ERESTARTSYS;
                         goto end;
                 }

                 // Whatever happen, we can read back the status to users mess
                 //args[0] = read_devfeature(priv);

                 status = read_status(priv);
                 if ( ATASTS_GOT_ERR(status) )
                         ret = -EIO;

                 // Not sure for arg0
                 args[0] = status;
                 args[1] = read_devfeature(priv);
                 args[2] = read_sectorcnt(priv);
                 args[3] = read_sectornum(priv);

                 NPRINTK("arg cmd=0x%2.2x, [0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x]\n", cmd, args[0], args[1], args[2], args[3]);
                 NPRINTK("return %d\n", ret);

                 if(ret>0)
                         ret = 0;

         end:
                 priv->io_inprogress = 0;
                 priv->ata_handler = ata_void_handler;
                 return ret;
         }


EXPORT_SYMBOL(mpc52xx_ata_dodrivecmd);
