/*
 *  mpc52xx_atablock / hwmisc.c
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */

/*
 * misc hw func (pio mode ...)
*/

/**************/
/**************/
/**************/
#include "mpc52xx_blockata.h"

/**************/
/**************/
/**************/
static int ata_doidentify_atapi(struct mpc52xx_blockata_priv *priv)
{
        u16 * drive_idendify;
        int ret;
        u16 OneShort;

        NPRINTK("\n");

        ret = 0;
        drive_idendify = priv->drive_identify;

        NPRINTK("id[00]=0x%2.2x\n", drive_idendify[0]);
        NPRINTK("id[85]=0x%2.2x\n", drive_idendify[85]);

        OneShort = drive_idendify[0];
        if ( (OneShort & 0xc000) != 0x8000)
                return -1;

        if (OneShort & (1<<7) )
                priv->IsRemovable = 1;
        else
                priv->IsRemovable = 0;

        priv->drive_canlba = 0;
        priv->IsATAPI = 1;

        priv->drive_identify_valid = 1;

        return ret;
}

static int ata_doidentify(struct mpc52xx_blockata_priv *priv)
{
        int ret;

        priv->drive_identify_valid = 0;

        ret = wait_ready(priv);
        if (ret!=0)
                goto error;

        write_cmd(priv, priv->UsePacket ? ATA_CMD_ID_ATAPI : ATA_CMD_ID_ATA);
        ret = wait_drq(priv);
        if (ret!=0)
                goto error;

        ret = mpc52xx_ata_docpupollread(priv, priv->drive_identify, 256);
        if (ret<0)
                goto error;

        priv->drive_identify[49] = priv->drive_identify[49] & ~(1<<8);

        memcpy(priv->drive_model, (char *) & priv->drive_identify[27], 40);
        priv->drive_model[39] = '\0';
        memcpy(priv->drive_firmware, (char *) & priv->drive_identify[23], 8);
        priv->drive_firmware[7] = '\0';

        if (priv->UsePacket)
                return ata_doidentify_atapi(priv);

        priv->drive_heads = priv->drive_identify[3];
        priv->drive_cylinders = priv->drive_identify[1];
        priv->drive_sectorspertrack = priv->drive_identify[6];
        priv->drive_cap = priv->drive_identify[49];

        priv->drive_canlba = priv->drive_cap & CAPF_LBA;
        priv->drive_canlba48 = ((priv->drive_identify[83]) & (1 << 10))  == (1 << 10) ;

        if (priv->drive_canlba) {
                priv->drive_sectors = (sector_t)
                                      ( (priv->drive_identify[60] & 0xffff)
                                        | ( (priv->drive_identify[61] << 16) & 0xffff0000) );
        } else {
                priv->drive_sectors = (sector_t)
                                      ( (priv->drive_identify[57] & 0xffff)
                                        | ( (priv->drive_identify[58] << 16) & 0xffff0000) );
        }


        NPRINTK("%s: id[0]=%x\n", __func__, priv->drive_identify[0]);
        NPRINTK("%s: id[1]=%x\n", __func__, priv->drive_identify[1]);
        NPRINTK("%s: id[3]=%x\n", __func__, priv->drive_identify[3]);
        NPRINTK("%s: id[6]=%x\n", __func__, priv->drive_identify[6]);
        NPRINTK("%s: id[53]=%x\n", __func__, priv->drive_identify[53]);
        NPRINTK("%s: id[54]=%x\n", __func__, priv->drive_identify[54]);
        NPRINTK("%s: id[55]=%x\n", __func__, priv->drive_identify[55]);
        NPRINTK("%s: id[56]=%x\n", __func__, priv->drive_identify[56]);

        NPRINTK("%s: id[57/58]=%x %x\n", __func__, priv->drive_identify[57], priv->drive_identify[58]);
        NPRINTK("%s: id[60/61]=%x %x\n", __func__, priv->drive_identify[57], priv->drive_identify[58]);
        NPRINTK("%s: id[82]=%x\n", __func__, priv->drive_identify[82]);
        NPRINTK("%s: id[83]=%x\n", __func__, priv->drive_identify[83]);


        NPRINTK("%s: model=%s, fw=%s. %d heads, %d cylinders, %lld sectors. LBA=%d\n",
                __func__,
                priv->drive_model, priv->drive_firmware,
                priv->drive_heads, priv->drive_cylinders, priv->drive_sectors, priv->drive_canlba );

        priv->drive_identify_valid = 1;
        return 0;

error:
        return -1;
}

int mpc52xx_ata_doidentify(struct mpc52xx_blockata_priv *priv)
{
        return ata_doidentify(priv);
}

/**************/
/**************/
int mpc52xx_ata_setpiomode(struct mpc52xx_blockata_priv *priv, int pio_mode)
{
        u8 setfeature_val;
        int ret;

        NPRINTK("%s: pio_mode=%d\n", __func__, pio_mode);

        switch(pio_mode) {
        case 0:
                setfeature_val = XFER_PIO_0;
                break;
        case 1:
                setfeature_val = XFER_PIO_1;
                break;
        case 2:
                setfeature_val = XFER_PIO_2;
                break;
        case 3:
                setfeature_val = XFER_PIO_3;
                break;
        case 4:
                setfeature_val = XFER_PIO_4;
                break;
        default:
                return -1;
        }

        write_sectorcnt(priv, setfeature_val);
        write_devfeature(priv, SETFEATURES_XFER);

        ret = wait_ready(priv);
        if (ret !=0)
                return ret;

        write_cmd(priv, ATA_CMD_SET_FEATURES);

        return wait_not_busy(priv);
}

/*
 * if pio_arg is <0 then, return the best pio mode available
*/
int mpc52xx_ata_isthispiovalid(struct mpc52xx_blockata_priv *priv, int pio_arg)
{
        int ret;
        int mode;

        u16 id_fieldval;
        u16 id_advpio;

        NPRINTK("pio_arg=%d\n",  pio_arg);

        if (!priv->drive_identify_valid) {
                ret = ata_doidentify(priv);
                if (ret!=0)
                        return -1;
        }

        id_advpio = priv->drive_identify[64];
        id_fieldval = priv->drive_identify[53];

        NPRINTK("id_advpio=%x, id_fieldval=%x\n",id_advpio,  id_fieldval);

        if (pio_arg<0) {
                pio_arg = 2;

                if (id_fieldval & 0x0002) {
                        if (id_advpio&0x2)
                                pio_arg = 4;
                        else if (id_advpio&0x1)
                                pio_arg = 3;
                }
        }

        NPRINTK("pio_arg=%d\n", pio_arg);

        switch(pio_arg) {
        case 0:
        case 1:
        case 2:
                mode = pio_arg;
                break;

        case 3:
                mode = ( (id_fieldval&0x0002) && (id_advpio&0x1) )? pio_arg : -1;
                break;

        case 4:
                mode = ( (id_fieldval&0x0002) && (id_advpio&0x2) )? pio_arg : -1;
                break;

        default:
                mode = -1;
        }

        return mode;
}


/**************/
#define IDENTIFY_MULTI_MAXSECMASK   0x00ff
#define IDENTIFY_MULTI_OKFLAG    0x00100

int mpx52xx_ata_dosetmulti(struct mpc52xx_blockata_priv *priv, u16 val)
{
        int ret;

        NPRINTK("set multi to %d\n", val);

        if (priv->UsePacket)
                return -1;

        write_sectorcnt(priv, val&0xff);

        ret = wait_ready(priv);
        if (ret !=0)
                return -1;

        ATA_DUMPREG
        write_cmd(priv, ATA_CMD_SETMULTI);

        ata_sleep(priv);
        ret = wait_not_busy(priv);
        if (ret !=0)
                return -1;

        return 0;
}

#define IS_MULTIOK(__identify_multi__)  \
(  \
 (__identify_multi__ & IDENTIFY_MULTI_OKFLAG)  \
  && (__identify_multi__ & IDENTIFY_MULTI_MAXSECMASK) \
)

static int ata_multi_probeandset(struct mpc52xx_blockata_priv *priv)
{
        int ret;
        u16 identify_multi;
        u16 identify_multimax;

        priv->multi_available = 0;

        if (priv->UsePacket)
                return -1;

        NPRINTK("id[59]=0x%x, id[47]=0x%x\n", priv->drive_identify[59], priv->drive_identify[47]);

        identify_multi = priv->drive_identify[59];
        if ( IS_MULTIOK(identify_multi) ) {
                priv->multi_available = 1;
                priv->multi_secpercmd = identify_multi & IDENTIFY_MULTI_MAXSECMASK;
        } else {
                identify_multimax = priv->drive_identify[47] & IDENTIFY_MULTI_MAXSECMASK;
                ret = mpx52xx_ata_dosetmulti(priv, identify_multimax);
                if (ret<0)
                        goto error;

                ret = ata_doidentify(priv);
                if (ret!=0)
                        goto error;

                NPRINTK("idenditify updated, identify_multimax=%d, s=%x, id[59]=0x%x, id[47]=0x%x\n",
                        identify_multimax, read_altstatus(priv), priv->drive_identify[59], priv->drive_identify[47]);

                identify_multi = priv->drive_identify[59];

                if ( IS_MULTIOK(identify_multi) ) {
                        priv->multi_available = 1;
                        priv->multi_secpercmd = identify_multi & IDENTIFY_MULTI_MAXSECMASK;
                } else {
                        priv->multi_secpercmd = MAX_SECPERREQ;
                }

        }

        return 0;

error:
        priv->multi_secpercmd = MAX_SECPERREQ;
        priv->multi_available = 0;

        return ret;
}


/**************/
static int ata_dodiag(struct mpc52xx_blockata_priv *priv)
{
        int ret;
        u8 diagcode;
        u8 sectorcount;
        u8 lbah;
        u8 lbal;
        u8 lbam;

        NPRINTK("1\n");

        ret = wait_not_busy(priv);
        if (ret!=0)
                goto error;

        NPRINTK("cmd\n");
        write_cmd(priv, ATA_CMD_EDD);

        wait_not_busy(priv);
        diagcode = read_devfeature(priv);

        NPRINTK("diagcode=%x\n", diagcode);

        /* Check if this device implement the PACKET feature set
         * see ATAPI6 spec section 9.12
         * Non Packet has this signature: 
         *   Sector Count 01h
         *   LBA Low     01h
         *   LBA Mid    00h
         *   LBA High    00h
         *   Device    00h
         *  
         * Packet has this signature: 
         *   Sector Count 01h
         *   LBA Low     01h
         *   LBA Mid    14h
         *   LBA High    EBh
         *
         *  
        */
        sectorcount = read_sectorcnt(priv);
        lbal = read_sectornum(priv);
        lbam = read_cyllo(priv);
        lbah = read_cylhi(priv);

        NPRINTK("sc=%2.2x lbal=%2.2x lbam=%2.2x lbah=%2.2x\n", sectorcount, lbal, lbam, lbah);

        if ( (sectorcount != 0x01) && (lbal != 0x01) )
                goto error;

        if ( (lbam==0) && (lbah==0) )
                priv->UsePacket = 0;
        else if ((lbam==0x14) && (lbah==0xeb) )
                priv->UsePacket = 1;
        else
                goto error;

        return ! (diagcode & 0x01);

error:
        NPRINTK("error %d\n", ret);
        ATA_DUMPREG

        return -1;
}
/**************/
void mpc52xx_ata_dumpreg(struct mpc52xx_blockata_priv *priv)
{
        u8 seccnt, sector, cyl_lo, cyl_hi, error, status;

        seccnt = read_sectorcnt(priv);
        sector = read_sectornum(priv);
        cyl_lo = read_cyllo(priv);
        cyl_hi = read_cylhi(priv);
        error  = read_error(priv);
        status = read_altstatus(priv);

        VPRINTK("seccnt=%x sector=%x, cyl_lo=%x, cyl_hi=%x, error=%x, status =%x \n",
                seccnt, sector, cyl_lo, cyl_hi, error, status);
}

static int ata_doreset(struct mpc52xx_blockata_priv *priv)
{
        int ret = 0;

        ret = wait_not_busy(priv);
        if (ret!=0)
                return -1;

        NPRINTK("set reset %x\n", read_altstatus(priv) );
        write_ctl(priv, ATA_SRST | ATA_NIEN);
        ata_sleep(priv);

        NPRINTK("release reset, %x\n", read_altstatus(priv));
        write_ctl(priv, ATA_NIEN);

        ata_sleep(priv);

        ret = wait_not_busy(priv);
        if (ret<0)
                return ret;

        write_devhead(priv, 0x00);
        ata_sleep(priv);

        wait_not_busy(priv);
        if (ret<0)
                return ret;

        return ATASTS_GOT_ERR( read_altstatus(priv) );
}

/**************/
static int ata_setupchs(struct mpc52xx_blockata_priv *priv)
{
        int ret = 0;

        NPRINTK("s=0x%2.2x, lba=%d, secper=%d, heads=%dn",
                read_altstatus(priv),
                priv->drive_canlba,
                priv->drive_sectorspertrack,
                priv->drive_heads);

        priv->drive_chs_ok = 0;

        if
        (
                (priv->drive_sectorspertrack == 0)
                || priv->drive_canlba
        )
		return -1;

        if (!priv->drive_identify_valid) {
                ret = ata_doidentify(priv);
                if (ret!=0)
                        return -1;
        }

        NPRINTK("wait1 s=%2.2x\n", read_altstatus(priv) );
        ret = wait_not_busy(priv);
        if (ret<0)
                return ret;

        write_sectorcnt(priv, priv->drive_sectorspertrack );
        write_devhead(priv, priv->drive_heads - 1);

        NPRINTK("wait2 s=%2.2x\n", read_altstatus(priv) );
        ret = wait_ready(priv);
        if (ret<0) {
                ATA_DUMPREG
                return ret;
        }

        write_cmd(priv, ATA_CMD_INIT_DEV_PARAMS);

        NPRINTK("wait3 s=%2.2x\n", read_altstatus(priv) );
        ret = wait_not_busy(priv);
        if (ret<0) {
                ATA_DUMPREG
                return ret;
        }

        ret = ata_doidentify(priv);
        if (ret!=0)
                return ret;

        if (!(priv->drive_identify[53] & 0x0001))
                return -1;

        priv->drive_chs_ok = 1;

        priv->drive_chs_heads = priv->drive_identify[54];
        priv->drive_chs_cylinders = priv->drive_identify[55];
        priv->drive_chs_sectorspertrack = priv->drive_identify[56];

        NPRINTK("ok. CHS=%d/%d/%ds=0x%2.2x\n",
                read_altstatus(priv),
                priv->drive_chs_cylinders,
                priv->drive_chs_heads,
                priv->drive_chs_sectorspertrack);

        return 0;
}

/**************/
/**************/
/**************/

/*
 * Simple and dump drive init func
 * reset, sanity check, drive presence, get drive info
*/
int mpc52xx_ata_init_drive(struct mpc52xx_blockata_priv *priv)
{
        int ret;

        ret = wait_not_busy(priv);
        if (ret!=0)
                goto error;

        write_ctl(priv, ATA_NIEN);

        NPRINTK("select dev 0\n");
        write_devhead(priv, 0x00);

        ata_sleep(priv);

        ata_doreset(priv);

        /* Some check ... */
        NPRINTK("regcheck %x\n", read_altstatus(priv));
        ret = mpc52xx_ata_regcheck(priv);
        if (ret!=0)
                goto error;

        /* diagnostic */
        NPRINTK("diag %x\n", read_altstatus(priv));
        ret = ata_dodiag(priv);
        if (ret!=0)
                goto error;

        /* Get information from the drive */
        NPRINTK("do id %x, UsePacket=%d\n", read_altstatus(priv),  priv->UsePacket);
        if (!priv->drive_identify_valid) {
                ret = ata_doidentify(priv);
                if (ret!=0)
                        goto error;
        }

        ata_setupchs(priv);
        ata_multi_probeandset(priv);

        return ret;

error:
        NPRINTK("%s: error %d, altsts=%x, err=%x\n", __func__, ret, read_altstatus(priv),  read_error(priv));
        ATA_DUMPREG;

        return -1;
}

u32 mpc52xx_ata_getregisterbase(struct mpc52xx_blockata_priv *priv)
{ return 0xf0003a00; }

/*
 * Export stuff
*/

EXPORT_SYMBOL(mpc52xx_ata_init_drive);
EXPORT_SYMBOL(mpc52xx_ata_getregisterbase);
EXPORT_SYMBOL(mpc52xx_ata_doidentify);
EXPORT_SYMBOL(mpc52xx_ata_setpiomode);

EXPORT_SYMBOL(mpc52xx_ata_isthispiovalid);
EXPORT_SYMBOL(mpx52xx_ata_dosetmulti);
