/*
 *  mpc52xx_blockata.h
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */


#ifndef __MPC52xx_BLOCKATA_H__
#define __MPC52xx_BLOCKATA_H__

#include <linux/major.h>
#include <linux/vmalloc.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/blkdev.h>
#include <linux/bitops.h>
#include <linux/genhd.h>
#include <linux/hdreg.h>
#include <linux/fs.h>

#include <asm/kmap_types.h>
#include <asm/mpc52xx.h>

/**************/
/**************/
/**************/
//#define __DEBUG__

/**************/
/**************/
/**************/
#include <platforms/efika/bestcomm.h>

#include "protos.h"
#include "mpc52xx_ide.h"


/**************/
/**************/
/**************/

#ifdef __DEBUG__
#define DEBUG
#define NPRINTK(fmt, args...) printk(KERN_ERR "%s: " fmt, __FUNCTION__, ## args)
//#define NPRINTK2(fmt, args...) printk(KERN_ERR "%s: " fmt, __FUNCTION__, ## args)
#define NPRINTK2(fmt, args...)

#define ATA_DUMPREG mpc52xx_ata_dumpreg(priv);

#else
#define NPRINTK(fmt, args...)
#define NPRINTK2(fmt, args...)
#define ATA_DUMPREG
#endif

/**************/
/**************/
/**************/

#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_VERBOSE
#define VPRINTK(fmt, args...) if ( printk_ratelimit() ) { printk(KERN_ERR DEVSKEL_DEVICENAME ": " fmt, ## args); }
#else
#define VPRINTK(fmt, args...)
#endif

#define DEVICE_USERLIMIT 1024
#define KERNEL_SECTOR_SIZE 512

// I notice 8 gives some good result
#define DRIVER_MAXHWSEG 8

/**************/
/**************/
/**************/


#ifndef ATA_CMD_RESET
#define ATA_CMD_RESET   0x08
#endif

#ifndef ATA_CMD_SETMULTI
#define ATA_CMD_SETMULTI   0xc6
#endif

/**************/
/**************/
/**************/

#ifndef bio_cur_sectors_idx
#define bio_cur_sectors_idx(__bio__,__idx__)    (bio_iovec_idx((__bio__), (__idx__))->bv_len >> 9)
#endif

#ifndef bio_getnext
#define bio_getnext(__bio__) ( (__bio__)->bi_next )
#endif


#ifndef bio_getsector
#define bio_getsector(__bio__) ( (__bio__)->bi_sector )
#endif

#ifndef bio_getcuridx
#define bio_getcuridx(__bio__) ( (__bio__)->bi_idx )
#endif

#ifndef bio_islastidx
#define bio_islastidx(__bio__, __idx__) ( (__idx__) >= (__bio__)->bi_vcnt )
#endif



/**************/
/**************/
/**************/

/*
 *
 * TO DO:
 * - better error handling
 * - Transaction timeout/reset
 * - test and fix LBA48 addressing
 * - code, test and fix CHS addressing
 *
 * History
 * - 0.1. 27.06.2006
 *        Initial revision
 *        PIO func, SDMA/CPU polling...
 *        Reliable operation (1 or 2 MB/s)
 *
 * - 0.2. 28.06.2006
 *        Interrupt based driver!
 *        IRQ handlers, handles better bio, bio vec, req...
 *        Reliable operation (write 1.5MB/s, read 3.5MB/s to 6.5MB/s)
 *
 * - 0.3b1. 29.06.2006
 *        Cleanup and small fixes
 *        Wait for DRQ interrupt instead of polling for read
 *        Add user count feature
 *
 * - 0.3b2. 30.06.2006
 *        Fixed the nopen/nrelease -> release was disabling IRQ even if
 *          there was still some people using it. I don't even speak about open!
 *        Cleanup and small fixes
 *
 * - 0.4  02.07.2006
 *        Driver totaly cleaned up  and spread other severals sources file
 *        Implemented LBA48 but untested due to lack of drive (you've been warmed!)
 *        
 * - 0.5  07.07.2006
 *        General cleanup
 *        Better SDMA task configuration
 *        Try to adjust the block queue config
 *
 * - 0.6b1 30.07.2006
 *        General cleanup
 *        Add HIDIO_DRIVE_CMD support
 *        Fixed (a bit) spinlock_* 
 *        Primilary ATAPI/Packet support
 *
 * - 0.6  31.07.2006
 *        General cleanup
 *        Nicely failed when ATAPI/Packet command is needed
 *        Fixed HDIO_DRIVE_CMD/Identify
 *
 * - 0.7  18.10.2006
 *        General cleanup
 *        fix text layout for Linux patch
 *        First try to init the drive and then the Linux block system becase
 *          the block code BUG() on free
*/


/**************/
/**************/
/**************/

#define MAX_DMA_BUFFERS 4
#define MAX_DMA_BUFFER_SIZE 512*256

#define DEVSKEL_DRIVERVERSION "0.7"
#define DEVSKEL_DRIVERNAME "MPC52xx ATA/PIO"
#define DEVSKEL_DEVICENAME "mpc52xx_ata"

#define MAX_SECPERREQ DRIVER_MAXHWSEG


/**************/
/**************/
/**************/

#ifndef MPC52xx_ATA_OFFSET
#define MPC52xx_ATA_OFFSET (0x3a00)
#endif

#define ATAFIFO_BUSADDR ( (u32) 0xf0003a60 )


/**************/
/**************/
/**************/

/* Helper to compute timing parameters */
#define CALC_CLK_VALUE_UP(c,v) (((v) + c - 1) / c)

/**************/
/**************/
/**************/
/* Private structures used by the driver */
struct mpc52xx_ata_timings
{
        u32 pio1;
        u32 pio2;
        u32 mdma1;
        u32 mdma2;
        u32 udma1;
        u32 udma2;
        u32 udma3;
        u32 udma4;
        u32 udma5;
        int using_udma;
};

/**************/
/**************/
/**************/
struct mpc52xx_blockata_priv
{
        struct gendisk *device_gendisk;
        struct request_queue *device_queue;
        int major;

        spinlock_t lock;

	wait_queue_head_t my_waitqueue;

	int drive_inited;
	int usercnt;

	int drive_identify_valid;
	u16 drive_identify[256];

        char drive_model[40];
        char drive_firmware[8];

        sector_t drive_sectors;
        int drive_sectorspertrack;
        int drive_cylinders;
        int drive_heads;

        int drive_chs_ok;
        int drive_chs_sectorspertrack;
        int drive_chs_cylinders;
        int drive_chs_heads;

#define CAPF_LBA (1<<9)
        u16 drive_cap;
        int drive_canlba;
        int drive_canlba48;

        u8 curio_atacmd;

        int io_inprogress;

        unsigned int ipb_period;	/* in ps */

        struct mpc52xx_ata __iomem *ata_regs;
        u32 ata_regs_bus;
        struct mpc52xx_ata_timings  timings[2];

        struct bestcomm_taskhandle taskhandle;
        struct bestcomm_taskhandle *sdma;

        int ata_irq;
        int sdma_irq;

        int multi_secpercmd;
        int multi_available;

        irqreturn_t (*sdma_handler) (struct mpc52xx_blockata_priv  *priv);
        irqreturn_t (*ata_handler) (struct mpc52xx_blockata_priv  *priv, u8 ata_status);

        int curio_sectodo;

        int curio_secidx;
        int curio_secpershot;
        u16 *curio_buffer;

        struct request *curio_req;
        struct bio *curio_bio;
        int curio_bioidx;
        sector_t curio_sector;

        int piomode;
 
        int UsePacket;
        int IsATAPI;
        int IsRemovable;
};


/**************/
/**************/
/**************/

#include "piofunc_inline.h"

#endif
