/*
 *  mpc52xx_atablock / piofunc_inline.h
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */


#ifndef __MPC52xx_BLOCKATA_PIOFUNC_H__
#define __MPC52xx_BLOCKATA_PIOFUNC_H__


/**************/
/**************/
/**************/

#include "mpc52xx_blockata.h"

/**************/
/**************/
/**************/

#include <asm/io.h>

/*
 * Pio helper
*/

static inline u8 read_altstatus(struct mpc52xx_blockata_priv *priv)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        return readb(&ata_regs->tf_control);
}

static inline u8 read_status(struct mpc52xx_blockata_priv *priv)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        return readb(&ata_regs->tf_command);
}

static inline u8 read_sectorcnt(struct mpc52xx_blockata_priv *priv)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        return readb(&ata_regs->tf_sec_count);
}

static inline u8 read_sectornum(struct mpc52xx_blockata_priv *priv)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        return readb(&ata_regs->tf_sec_num);
}

static inline u8 read_cylhi(struct mpc52xx_blockata_priv *priv)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        return readb(&ata_regs->tf_cyl_high);
}

static inline u8 read_cyllo(struct mpc52xx_blockata_priv *priv)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        return readb(&ata_regs->tf_cyl_low);
}

static inline u8 read_devfeature(struct mpc52xx_blockata_priv *priv)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        return readb(&ata_regs->tf_features);
}
static inline u8 read_error(struct mpc52xx_blockata_priv *priv)
{
        return read_devfeature(priv);
}

static inline u16 read_data(struct mpc52xx_blockata_priv *priv)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        return readw(&ata_regs->tf_data);
}

/**************/
static inline void write_data(struct mpc52xx_blockata_priv *priv, u16 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writew(val, &ata_regs->tf_data);
}

static inline void write_devfeature(struct mpc52xx_blockata_priv *priv, u8 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writeb(val, &ata_regs->tf_features);
}


static inline void write_cyllo(struct mpc52xx_blockata_priv *priv, u8 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writeb(val, &ata_regs->tf_cyl_low);
}

static inline void write_cylhi(struct mpc52xx_blockata_priv *priv, u8 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writeb(val, &ata_regs->tf_cyl_high);
}

static inline void write_cmd(struct mpc52xx_blockata_priv *priv, u8 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writeb(val, &ata_regs->tf_command);
}

static inline void write_ctl(struct mpc52xx_blockata_priv *priv, u8 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writeb(val, &ata_regs->tf_control);
}

static inline void write_sectornum(struct mpc52xx_blockata_priv *priv, u8 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writeb(val, &ata_regs->tf_sec_num);
}

static inline void write_sectorcnt(struct mpc52xx_blockata_priv *priv, u8 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writeb(val, &ata_regs->tf_sec_count);
}


static inline void write_devhead(struct mpc52xx_blockata_priv *priv, u8 val)
{
        struct mpc52xx_ata __iomem *ata_regs = priv->ata_regs;
        writeb(val, &ata_regs->tf_dev_head);
}


/**************/
static inline void ata_sleep(struct mpc52xx_blockata_priv *priv)
{
        read_altstatus(priv);
        udelay(500);
}

static inline void ata_sleep2(struct mpc52xx_blockata_priv *priv)
{
        read_altstatus(priv);
        udelay(1);
}

/**************/
#define ATASTS_IS_NOTBUSY(__status__)  ( ! ( (__status__) & ATA_BUSY) )
#define ATASTS_IS_READY(__status__) ( ( (__status__) & (ATA_DRDY|ATA_BUSY) ) == ATA_DRDY )
#define ATASTS_IS_DRQ(__status__) ( (__status__) & ATA_DRQ)
#define ATASTS_GOT_ERR(__status__) ( (__status__) & ATA_ERR)

#define WAIT_TIMEOUT  (1000*1000)


static inline u8 read_mystatus(struct mpc52xx_blockata_priv *priv)
{
        return read_altstatus(priv);
}

static inline int wait_not_busy(struct mpc52xx_blockata_priv *priv)
{
        int timeout;
        u8 status;

        timeout = WAIT_TIMEOUT;

        status = read_mystatus(priv);
        if ( ATASTS_IS_NOTBUSY(status) )
                goto end;

        while(timeout--) {
                status = read_mystatus(priv);
                if (  ATASTS_IS_NOTBUSY(status)  )
                        goto end;

                ata_sleep2(priv);
        }

        NPRINTK("%s: timeout, %x\n", __func__, status);

        return -1;

end:
        return 0;
}


static inline int wait_ready(struct mpc52xx_blockata_priv *priv)
{
        int timeout;
        u8 status;

        timeout = WAIT_TIMEOUT;

        status = read_mystatus(priv);
        if ( ATASTS_IS_READY(status) )
                goto end;

        while(timeout--) {
                status = read_mystatus(priv);
                if (  ATASTS_IS_READY(status)  )
                        goto end;

                ata_sleep2(priv);
        }

        NPRINTK("%s: timeout, %x\n", __func__, status);

        return -1;

end:
        return ATASTS_GOT_ERR(status) ? -1 : 0;
}

static inline int wait_drq(struct mpc52xx_blockata_priv *priv)
{
        int timeout;
        u8 status;

        timeout = WAIT_TIMEOUT;

        status = read_mystatus(priv);
        if ( ATASTS_IS_DRQ(status) )
                goto end;

        while(timeout--) {
                status = read_mystatus(priv);
                if ( ATASTS_IS_DRQ(status) )
                        goto end;

                ata_sleep2(priv);
        }

        NPRINTK("%s: timeout, %x\n", __func__, status);
        return -1;

end:
        return ATASTS_GOT_ERR(status) ? -1 : 0;
}

#endif
