/*
 *  protos.h
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */


/*
 * Simply summaries every protos we use over the driver
*/

#ifndef __MPC52xx_BLOCKATA_PROTOS_H__
#define __MPC52xx_BLOCKATA_PROTOS_H__

#include <linux/irq.h>
#include <linux/module.h>
#include <linux/interrupt.h>

struct mpc52xx_blockata_priv;

/**********/
/* HWMISC */
/**********/
int mpc52xx_ata_setpiomode(struct mpc52xx_blockata_priv *priv, int pio_mode);
int mpc52xx_ata_init_drive(struct mpc52xx_blockata_priv *priv);
int mpc52xx_ata_isthispiovalid(struct mpc52xx_blockata_priv *priv, int pio_arg);
int mpc52xx_ata_doidentify(struct mpc52xx_blockata_priv *priv);
u32 mpc52xx_ata_getregisterbase(struct mpc52xx_blockata_priv *priv);
int mpx52xx_ata_dosetmulti(struct mpc52xx_blockata_priv *priv, u16 val);


/**********/
/*  ATA   */
/**********/
int mpc52xx_ata_setupsector(struct mpc52xx_blockata_priv  *priv, sector_t sector, int sector_num, int is_write);
int mpc52xx_ata_isthispiovalid(struct mpc52xx_blockata_priv  *priv, int pio_arg);
void mpc52xx_ata_dumpreg(struct mpc52xx_blockata_priv  *priv);
int mpc52xx_ata_doreset(struct mpc52xx_blockata_priv  *priv);
int mpc52xx_ata_regcheck(struct mpc52xx_blockata_priv  *priv);
int mpc52xx_ata_dodrivereset(struct mpc52xx_blockata_priv  *priv);

/************/
/*DODRIVECMD*/
/************/
int mpc52xx_ata_dodrivecmd(
        struct mpc52xx_blockata_priv *priv,
        unsigned long *irqflags,
        u8 *arg);

#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA


/**********/
/*  SDMA  */
/**********/
int mpc52xx_ata_sdma_setup(struct mpc52xx_blockata_priv  *priv);

void sdma_ata_rx_init(struct bestcomm_taskhandle *mytaskhandle);
void sdma_ata_tx_init(struct bestcomm_taskhandle *mytaskhandle);
void sdma_ata_reset(struct bestcomm_taskhandle *mytaskhandle);
int sdma_ata_getirq(struct bestcomm_taskhandle *mytaskhandle);
void sdma_ata_clear_irq(struct bestcomm_taskhandle *mytaskhandle);
void sdma_ata_disable(struct bestcomm_taskhandle *mytaskhandle);
void sdma_ata_enable(struct bestcomm_taskhandle *mytaskhandle);

void sdma_ata_submit_buffer(
        struct bestcomm_taskhandle *mytaskhandle,
        void *cookie,
        void *data1,
        void *data2,
        int length);

#endif /* CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA */

/**********/
/*TRANSFER*/
/**********/
int mpc52xx_ata_docpupollread(
        struct mpc52xx_blockata_priv *priv,
        void *buffer,
        int len);

int mpc52xx_ata_docpupollwrite(
        struct mpc52xx_blockata_priv*priv,
        void *buffer,
        int len);

int mpc52xx_ata_dotransfer(
        struct mpc52xx_blockata_priv *priv,
        struct request *req,
        struct bio *bio,
        int bio_index,
        sector_t sector,
        int sectorcnt,
        char *buffer,
        int is_write);

irqreturn_t sdma_void_handler(struct mpc52xx_blockata_priv *priv);
irqreturn_t ata_void_handler(struct mpc52xx_blockata_priv *priv, u8 ata_status);

void mpc52xx_ata_ack_blkreq(struct mpc52xx_blockata_priv *priv, int retval);

#endif
