/*
 *  mpc52xx_atablock / piofunc.c
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */


/**************/
/**************/
/**************/

#include "mpc52xx_blockata.h"


/**************/
/**************/
/**************/


#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA

#include <asm/prom.h>

/**************/
/**************/
/**************/

/**************/
void sdma_ata_rx_init(struct bestcomm_taskhandle *mytaskhandle)
{
        struct sdma_ata_inc *inc;
        inc = (struct sdma_ata_inc *)bestcomm_taskget_inctable(mytaskhandle);

        NPRINTK("inc = %p\n", inc);

        inc->incr_bytes = -(s16)sizeof(u16);
        inc->incr_src = 0;
        inc->incr_dst = sizeof(u16);
}

/**************/
/*
 * Initialize ATA transmit task.
 */
void sdma_ata_tx_init(struct bestcomm_taskhandle *mytaskhandle)
{
        struct sdma_ata_inc *inc;
        inc = (struct sdma_ata_inc *)bestcomm_taskget_inctable(mytaskhandle);

        NPRINTK("inc = %p\n", inc);

        inc->incr_bytes = -(s16)sizeof(u16);
        inc->incr_src = sizeof(u16);
        inc->incr_dst = 0;
}

/**************/
void sdma_ata_reset(struct bestcomm_taskhandle *mytaskhandle)
{
        struct sdma_ata_var *var;
        var = (struct sdma_ata_var *)bestcomm_taskget_vartable(mytaskhandle);

        NPRINTK("var = %p\n", var);

        sdma_reset_buffers2((struct sdma *) mytaskhandle);

        var->bd_start = var->bd_base;
}


/**************/
int sdma_ata_getirq(struct bestcomm_taskhandle *mytaskhandle)
{
        return bestcomm_taskget_irq(mytaskhandle);
}

/**************/
void sdma_ata_clear_irq(struct bestcomm_taskhandle *mytaskhandle)
{
        bestcomm_taskclear_irq(mytaskhandle);
}

/**************/
void sdma_ata_disable(struct bestcomm_taskhandle *mytaskhandle)
{
        bestcomm_taskdisable(mytaskhandle);
}

/**************/
void sdma_ata_enable(struct bestcomm_taskhandle *mytaskhandle)
{
        bestcomm_taskenable(mytaskhandle);
}

int mpc52xx_ata_sdma_setup(struct mpc52xx_blockata_priv  *priv)
{
        struct sdma *s;

        s = sdma_ata_preinit(MAX_DMA_BUFFERS);
        if (!s)
                return -1;

        priv->sdma = (struct bestcomm_taskhandle *)s;
        return sdma_ata_init((struct bestcomm_taskhandle *)s, MAX_DMA_BUFFER_SIZE);
}

/**************/
void sdma_ata_submit_buffer(
        struct bestcomm_taskhandle *mytaskhandle,
        void *cookie,
        void *data1,
        void *data2,
        int length)
{
        NPRINTK(" d1=%p, d2=%p, len=%d\n",  data1, data2, length);

        sdma_submit_buffer2(
                (struct sdma *)mytaskhandle,
                cookie, data1, data2, length);
}

/*
 * Export stuff
*/

EXPORT_SYMBOL(sdma_ata_rx_init);
EXPORT_SYMBOL(sdma_ata_tx_init);
EXPORT_SYMBOL(sdma_ata_disable);
EXPORT_SYMBOL(sdma_ata_enable);
EXPORT_SYMBOL(sdma_ata_submit_buffer);
EXPORT_SYMBOL(sdma_ata_reset);
EXPORT_SYMBOL(sdma_ata_getirq);
EXPORT_SYMBOL(sdma_ata_clear_irq);
EXPORT_SYMBOL(mpc52xx_ata_sdma_setup);


#endif // CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA

