/*
 *  mpc52xx_atablock / transfer.c
 *
 *  Copyright 2006 bplan GmbH
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file "COPYING" in the main directory of this archive
 *  for more details.
 */

/*
 * real brain!
*/


#include "mpc52xx_blockata.h"


/**************/
/**************/
/**************/
static int ata_dopollwrite(
        struct mpc52xx_blockata_priv *priv,
        void *buffer,
        int len);
static int ata_dopollread(
        struct mpc52xx_blockata_priv *priv,
        void *buffer,
        int len);

/**************/
/**************/
/**************/
/*
 * We could check for unexpected interrupt or status here
*/

irqreturn_t sdma_void_handler(struct mpc52xx_blockata_priv *priv)
         {
                 NPRINTK("%s: \n", __func__);
                 return IRQ_HANDLED;
         }

         irqreturn_t ata_void_handler(struct mpc52xx_blockata_priv *priv, u8 ata_status)
                  {
                          NPRINTK("%s: status=0x%2.2x\n", __func__, ata_status);
                          return IRQ_HANDLED;
                  }



#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA

                  /**************/
                  static int start_writerequest(
                          struct mpc52xx_blockata_priv *priv,
                          struct request *req,
                          struct bio *bio,
                          int bio_index,
                          sector_t sector,
                          int sectorcnt,
                          u16 * buffer);
static void submit_writebuffer(
        struct mpc52xx_blockata_priv *priv,
        u16 *buffer,
        int curio_secidx);

static void submit_readbuffer(
        struct mpc52xx_blockata_priv *priv,
        u16 *buffer,
        int curio_secidx);
static int submitandenable_readbuffer(
        struct mpc52xx_blockata_priv *priv,
        u16 *buffer);
static int start_readrequest(
        struct mpc52xx_blockata_priv *priv,
        struct request *req,
        struct bio *bio,
        int bio_index,
        sector_t sector,
        int sectorcnt,
        u16 * buffer);

/**************/
void mpc52xx_ata_ack_blkreq(struct mpc52xx_blockata_priv *priv, int retval)
{
        struct request *req;

        req = priv->curio_req;

        priv->curio_bio = NULL;
        priv->curio_req = NULL;
        priv->sdma_handler = sdma_void_handler;
        priv->ata_handler = ata_void_handler;

        priv->io_inprogress = 0;

        // Ack the req if any
        if (req)
                end_request(req, retval);

        // Make sure to restart the queue
        blk_start_queue(priv->device_queue);
}

/**************/
static int inhandlercheck_atastatus(
        struct mpc52xx_blockata_priv *priv,
        u8 ata_status)
{
        if ( ATASTS_GOT_ERR(ata_status) )
        {
                VPRINTK("ATA Error, transfer aborted");
                ATA_DUMPREG

                mpc52xx_ata_ack_blkreq(priv, 0);

                // taut!
                return -1;
        }

        return 0;
}


/**************/
/*
 * This func will do the necessary operation when a bio iovec is done
 * -> end the transfer if there are no more job
 * -> start a new io vec if any
 * -> start a new bio if any
 *
 * This is the same for RX and TX buffer that's why I wrote this static
 * (and hopefully inlined !) func
 * 
*/
static inline u16 *handle_nextbio(
        struct mpc52xx_blockata_priv *priv,
        struct request *req,
        struct bio **ptr_bio,
        int *ptr_bio_idx,
        sector_t *ptr_sector,
        int *ptr_sectorcnt)
{
        struct bio *bio;
        int bio_idx;
        int lastvec_sectorcnt;
        sector_t sector;

        bio = priv->curio_bio;

        // need to do!
        NPRINTK2(KERN_DEBUG "One iovec done. bio idx=%d. cnt=%d\n", bio->bi_idx, bio->bi_vcnt);

        /* Unmap the previous one
           * kmap stuff is a bit strange
           * the LDD manual pass the bio, but the 
           * macro seen to accept the buffer...
           * maybe API change between the LDD (3rd release)
           * and this Linux kernel (2.6.16.15 but also in 2.6.17.1).
           * anyway, kmap* seens void.
        */
        __bio_kunmap_atomic( bio_data(bio), KM_USER0);

        bio_idx = priv->curio_bioidx;
        lastvec_sectorcnt = bio_cur_sectors_idx(bio, bio_idx);

        bio_idx++;
        if (bio_islastidx(bio, bio_idx) )
        {
                // Finish with this bio

                // Let's see if there is another bio in this req
                bio = bio_getnext(bio);
                if (!bio) {
                        // we have done!
                        NPRINTK(KERN_DEBUG "req (%p, bio=%p) over, ack it and start the queue\n",
                                req, bio);

                        mpc52xx_ata_ack_blkreq(priv, 1);

                        // vertig
                        return NULL;
                }

                NPRINTK(KERN_DEBUG "bio over,but new one %p\n", bio);

                // Set up stuff for the new bio
                bio_idx = bio_getcuridx(bio);
                sector = bio_getsector(bio);
        } else
        {
                // Set up stuff for the next bio iovec
                NPRINTK(KERN_DEBUG "bio iovec over,but new one %d\n", bio_idx);

                sector = priv->curio_sector + lastvec_sectorcnt;
        }

        *ptr_bio = bio;
        *ptr_sector = sector;
        *ptr_bio_idx = bio_idx;
        *ptr_sectorcnt = bio_cur_sectors_idx(bio, bio_idx);

        // Small notes, I'm really not sure kmap atomix is smart here
        // We should prolly use kmap_irq as we will prolly be called from an interrupt
        return __bio_kmap_atomic(bio, bio_idx, KM_USER0);
}

/**************/

irqreturn_t ata_complete_txtransfer_handler(struct mpc52xx_blockata_priv *priv, u8 ata_status)
         {
                 int secidx;
                 int sectodo;
                 int secpershot;
                 u16 *buffer;

                 NPRINTK2("curio_secidx=%d, curio_sectodo=%d. bio=%p, req=%p. s=0x%x\n",
                          priv->curio_secidx, priv->curio_sectodo, priv->curio_bio,  priv->curio_req, ata_status );

                 // Make sure to kill the task first
                 sdma_ata_disable(priv->sdma);

                 // Sanity check!
                 if ( ! priv->io_inprogress )
                         goto end;

                 if ( inhandlercheck_atastatus(priv, ata_status ) )
                         goto end;

                 // grab some data
                 secidx = priv->curio_secidx;
                 sectodo = priv->curio_sectodo;
                 secpershot = priv->curio_secpershot;

                 secidx += secpershot;
                 if (secidx < sectodo)
                 {
                         if ( ATASTS_IS_READY(ata_status)  ) {
                                 // Update the buffer to the position of the new sector to write
                                 // buffer is a u16*
                                 buffer = ( (u16 *) priv->curio_buffer) + (priv->curio_secpershot * 256);

                                 submit_writebuffer(priv, buffer, secidx);

                                 sdma_ata_enable(priv->sdma);
                         } else {
                                 // Error here, DRQ should be set
                                 goto end;
                         }
                 } else
                 {
                         struct bio *bio;
                         struct request *req;
                         int ret;
                         int bio_idx;
                         int sectorcnt;
                         u16 *buffer;
                         sector_t sector;

                         req =  priv->curio_req;
                         buffer = handle_nextbio(priv, req, &bio, &bio_idx, &sector, &sectorcnt);

                         ret =
                                 buffer
                                 ? start_writerequest(priv, req, bio, bio_idx, sector, sectorcnt, buffer)
                                 : -1;
                 }

         end:
                 return IRQ_HANDLED;
         }



         /**************/
         irqreturn_t sdma_complete_rxtransfer_handler(struct mpc52xx_blockata_priv *priv);

irqreturn_t ata_wait_rxready_handler(struct mpc52xx_blockata_priv *priv, u8 ata_status)
         {
                 NPRINTK2("%s: \n", __func__);

                 // Sanity check!
                 if (!priv->io_inprogress)
                 {
                         // Should report some erro rhere!
                         goto end;
                 }

                 if ( inhandlercheck_atastatus(priv, ata_status ) )
                         goto end;

                 if ( ATASTS_IS_DRQ(ata_status) )
                 {
                         // Set up the new handlers
                         priv->sdma_handler = sdma_complete_rxtransfer_handler;
                         priv->ata_handler = ata_void_handler;

                         // Now start the transfer
                         submitandenable_readbuffer(priv, priv->curio_buffer);
                 }

         end:
                 return IRQ_HANDLED;
         }

         irqreturn_t sdma_complete_rxtransfer_handler(struct mpc52xx_blockata_priv *priv)
                  {
                          int secidx;
                          int sectodo;
                          int secpershot;
                          u16 *buffer;

                          NPRINTK("curio_secidx=%d, curio_sectodo=%d. bio=%p, req=%p.\n",
                                  priv->curio_secidx, priv->curio_sectodo, priv->curio_bio,  priv->curio_req  );

                          // Make sure to kill the task first
                          sdma_ata_disable(priv->sdma);

                          // Sanity check!
                          if ( ! priv->io_inprogress )
                                  goto end;

                          // grab
                          secidx = priv->curio_secidx;
                          sectodo = priv->curio_sectodo;
                          secpershot = priv->curio_secpershot;

                          if ( inhandlercheck_atastatus(priv, read_altstatus(priv) ) )
                                  goto end;

                          secidx += secpershot;
                          if (secidx < sectodo)
                          {
#if 0
                                  u8 ata_status;

                                  // Rx transfer for this sector over, wait for intr for the new one
                                  priv->sdma_handler = sdma_void_handler;
                                  priv->ata_handler = ata_wait_rxready_handler;

                                  // reading the status reg should acknowledge this transfer
                                  // and start the next  one
                                  ata_status = read_status(priv);
#else
                                  // Stuff data to transfer for this iovec
                                  //if ( IS_READY( read_altstatus(priv) )  )
                                  {
                                          buffer = ( (u16 *) priv->curio_buffer) + (secpershot * 256);

                                          NPRINTK("%s: ATA ready -> read task started. Sector=%lld\n", __func__, priv->curio_sector);

                                          submit_readbuffer(priv, buffer, secidx);

                                          // I don't like to wait here
                                          wait_drq(priv);
                                          sdma_ata_enable(priv->sdma);
                                  }
#endif
                          } else
                          {
                                  struct bio *bio;
                                  struct request *req;
                                  int ret;
                                  int bio_idx;
                                  int sectorcnt;
                                  u16 *buffer;
                                  sector_t sector;

                                  req = priv->curio_req;
                                  buffer = handle_nextbio(priv, req, &bio, &bio_idx, &sector, &sectorcnt);

                                  ret =
                                          buffer
                                          ? start_readrequest(priv, req, bio, bio_idx, sector, sectorcnt, buffer)
                                          : -1;
                          }

                  end:
                          return IRQ_HANDLED;
                  }


                  /*
                   * This func will setup the ATA and the SDMA buffers
                   * and install the ATA handler
                  */
                  /**************/
                  static void submit_writebuffer(
                          struct mpc52xx_blockata_priv *priv,
                          u16 *buffer,
                          int curio_secidx)
                  {
                          u32 port_BusAddr;
                          u32 addr_BusAddr;

                          port_BusAddr = ATAFIFO_BUSADDR;
                          addr_BusAddr = virt_to_phys( (void*) buffer);

                          NPRINTK("priv=%p, , port bus=0x%8.8x, addr bus =0x%8.8x, buf=%p, idx=%d\n",
                                  priv, port_BusAddr, addr_BusAddr, buffer, curio_secidx);

                          // reset the buf
                          sdma_ata_reset(priv->sdma);

                          sdma_ata_submit_buffer(
                                  priv->sdma,
                                  (void *) buffer,
                                  (void *) addr_BusAddr,
                                  (void *) port_BusAddr,
                                  512*priv->curio_secpershot);

                          priv->curio_secidx = curio_secidx;
                          priv->curio_buffer = buffer;
                  }


                  static int start_writerequest(
                          struct mpc52xx_blockata_priv *priv,
                          struct request *req,
                          struct bio *bio,
                          int bio_index,
                          sector_t sector,
                          int sectorcnt,
                          u16 *buffer)
                  {

                          int ret;
                          int secperreq;
                          int sectodo;
                          int secpershot;

                          NPRINTK("%s: sector=%ld, sectorcnt=%d, buffer=%p\n", __func__, (long) sector, sectorcnt,  buffer);

                          secperreq = priv->multi_secpercmd;
                          sectodo = sectorcnt > secperreq ? secperreq : sectorcnt;
                          secpershot = sectodo ;

                          ret = wait_not_busy(priv);
                          if (ret!=0)
                                  return ret;
                          mpc52xx_ata_setupsector(priv, sector, sectodo, 1);

                          // I don't like to call wait here
                          ret = wait_ready(priv);
                          if (ret!=0)
                          {
                                  NPRINTK("can't wait rdy=%d. 0x%x\n", ret, read_altstatus(priv));
                                  ATA_DUMPREG
                                  return ret;
                          }

                          priv->curio_sector = sector;
                          priv->curio_secpershot = secpershot;
                          priv->curio_sectodo = sectodo;
                          priv->curio_req = req;
                          priv->curio_bio = bio;
                          priv->curio_bioidx = bio_index;

                          priv->sdma_handler = sdma_void_handler;
                          priv->ata_handler =  ata_complete_txtransfer_handler;

                          priv->io_inprogress = 1;

                          write_cmd(priv, priv->curio_atacmd);
                          ret = wait_drq(priv);
                          if (ret!=0)
                          {
                                  ATA_DUMPREG
                                  return ret;
                          }

                          submit_writebuffer(priv, (u16 *) buffer, 0);
                          sdma_ata_enable(priv->sdma);

                          return 0;
                  }


                  /**************/

                  /**************/
                  static void submit_readbuffer(
                          struct mpc52xx_blockata_priv *priv,
                          u16 *buffer,
                          int curio_secidx)
                  {
                          u32 port_BusAddr;
                          u32 addr_BusAddr;

                          port_BusAddr = ATAFIFO_BUSADDR;
                          addr_BusAddr = virt_to_phys( (void*) buffer);

                          NPRINTK2("priv=%p, , port bus=0x%8.8x, addr bus =0x%8.8x, buf=%p, idx=%d\n",
                                   priv, port_BusAddr, addr_BusAddr, buffer, curio_secidx);

                          // reset the buf
                          sdma_ata_reset(priv->sdma);

                          sdma_ata_submit_buffer(
                                  priv->sdma,
                                  (void *) buffer,
                                  (void *) port_BusAddr,
                                  (void *) addr_BusAddr,
                                  512*priv->curio_secpershot);

                          priv->curio_secidx = curio_secidx;
                          priv->curio_buffer = buffer;
                  }

                  static int submitandenable_readbuffer(
                          struct mpc52xx_blockata_priv *priv,
                          u16 *buffer)
                  {

                          submit_readbuffer(priv, (u16 *) buffer, 0);
                          sdma_ata_enable(priv->sdma);

                          return 0;
                  }

                  static int start_readrequest(
                          struct mpc52xx_blockata_priv *priv,
                          struct request *req,
                          struct bio *bio,
                          int bio_index,
                          sector_t sector,
                          int sectorcnt,
                          u16 *buffer)
                  {
                          int ret;
                          int secperreq;
                          int sectodo;
                          int secpershot;

                          NPRINTK2("%s: sector=%ld, sectorcnt=%d, buffer=%p, cmd=0%2.2x\n", __func__, (long) sector, sectorcnt,  buffer, priv->curio_atacmd);

                          secperreq = priv->multi_secpercmd;
                          sectodo = sectorcnt > secperreq ? secperreq : sectorcnt;
                          secpershot = sectodo;

                          ret = wait_not_busy(priv);
                          if (ret!=0)
                                  return ret;

                          mpc52xx_ata_setupsector(priv, sector, sectodo, 0);

                          ret = wait_ready(priv);
                          if (ret!=0)
                          {
                                  NPRINTK("can't wait rdy=%d. 0x%x\n", ret, read_altstatus(priv));
                                  ATA_DUMPREG
                                  return ret;
                          }

                          priv->curio_sector = sector;
                          priv->curio_secpershot = secpershot;
                          priv->curio_sectodo = sectodo;
                          priv->curio_req = req;
                          priv->curio_bio = bio;
                          priv->curio_bioidx = bio_index;

                          priv->sdma_handler = sdma_void_handler;
                          priv->ata_handler = ata_wait_rxready_handler;

                          priv->io_inprogress = 1;

                          /*
                           * We only start the task when we got the drive int
                          */
                          priv->curio_secidx = 0;
                          priv->curio_buffer = buffer;
                          write_cmd(priv, priv->curio_atacmd);

                          return 0;
                  }
#endif



#ifndef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA
                  /**************/
                  /**************/
                  /**************/
                  /*
                   * This do_read func use polling -> not interrupt!
                  */

                  static int do_readrequest(
                          struct mpc52xx_blockata_priv *priv,
                          struct request *req,
                          struct bio *bio,
                          int bio_index,
                          sector_t sector,
                          int sectorcnt,
                          u16 * buffer)
                  {
                          int ret;
                          int secperreq;
                          int sectodo;
                          int sectorcnt_original = sectorcnt;

                          NPRINTK("%s: sector=%ld, sectorcnt=%d, buffer=%p\n", __func__, (long) sector, sectorcnt, buffer);

                          secperreq = priv->multi_secpercmd;

                          while(sectorcnt>0)
                          {
                                  int i;
                                  sectodo = sectorcnt > secperreq ? secperreq : sectorcnt;

                                  ret = wait_not_busy(priv);
                                  if (ret!=0)
                                          return -1;
                                  mpc52xx_ata_setupsector(priv, sector, sectodo, 0);

                                  ret = wait_ready(priv);
                                  if (ret!=0) {
                                          ATA_DUMPREG
                                          return ret;
                                  }

                                  if (priv->multi_available) {
                                          write_cmd(priv, ATA_CMD_READ_MULTI);
                                          ret = wait_drq(priv);
                                          if (ret!=0) {
                                                  ATA_DUMPREG
                                                  return ret;
                                          }
                                          ret = ata_dopollread(priv, buffer, 256*sectodo );
                                          buffer = ( ((u16*) buffer) + 256*sectodo );
                                  } else {
                                          write_cmd(priv, ATA_CMD_PIO_READ);

                                          for(i=0; i < sectodo; i++) {
                                                  ret = wait_drq(priv);
                                                  if (ret!=0) {
                                                          ATA_DUMPREG
                                                          return ret;
                                                  }

                                                  ret = ata_dopollread(priv, buffer, 256);

                                                  // Check!

                                                  buffer =( ((u16*) buffer) + 256 );
                                          }
                                  }

                                  sector += sectodo;
                                  sectorcnt -= sectodo;
                          }

                          return sectorcnt_original ;
                  }

                  /**************/
                  static int do_writerequest(
                          struct mpc52xx_blockata_priv *priv,
                          struct request *req,
                          struct bio *bio,
                          int bio_index,
                          sector_t sector,
                          int sectorcnt,
                          u16 *buffer)
                  {
                          int ret;
                          int secperreq;
                          int sectodo;
                          int sectorcnt_original = sectorcnt;

                          NPRINTK("%s: sector=%ld, sectorcnt=%d, buffer=%p\n", __func__, (long) sector, sectorcnt,  buffer);

                          secperreq = priv->multi_secpercmd;

                          while(sectorcnt)
                          {
                                  int i;
                                  sectodo = sectorcnt > secperreq ? secperreq : sectorcnt;

                                  ret = wait_not_busy(priv);
                                  if (ret!=0)
                                          return ret;
                                  mpc52xx_ata_setupsector(priv, sector, sectodo, 1);

                                  ret = wait_ready(priv);
                                  if (ret!=0) {
                                          NPRINTK("can't wait rdy=%d. 0x%x\n", ret, read_altstatus(priv));
                                          ATA_DUMPREG
                                          return ret;
                                  }

                                  if (priv->multi_available) {
                                          /* If the drive support the multi write, let's go!
                                           * should be really faster as we don't have to wait for DRQ
                                           * However, either I make something wrong, either I did not find 
                                           * any disk supporting this!
                                          */
                                          write_cmd(priv, ATA_CMD_WRITE_MULTI);
                                          ret = wait_drq(priv);
                                          if (ret!=0) {
                                                  NPRINTK("can't wait drq %d\n", ret);
                                                  ATA_DUMPREG
                                                  return ret;
                                          }

                                          ret = ata_dopollwrite(priv, buffer, 256*sectodo);
                                          buffer = ( (u16*) buffer + 256*sectodo );
                                  } else {
                                          write_cmd(priv, ATA_CMD_PIO_WRITE);
                                          for(i=0; i < sectodo; i++) {
                                                  ret = wait_drq(priv);
                                                  if (ret!=0) {
                                                          ATA_DUMPREG
                                                          return ret;
                                                  }

                                                  ret = ata_dopollwrite(priv, buffer, 256 );

                                                  // Check!

                                                  buffer = ( ((u16*) buffer) + 256 );
                                          }
                                  }

                                  sector += sectodo;
                                  sectorcnt-= sectodo;
                          }

                          //NPRINTK("return %d\n", sectorcnt_original);
                          return sectorcnt_original;
                  }
#endif /* CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA */

                  /**************/
                  /**************/
                  /**************/
                  /*
                   * This is call by nicoskel
                  */

                  int mpc52xx_ata_dotransfer(
                          struct mpc52xx_blockata_priv *priv,
                          struct request *req,
                          struct bio *bio,
                          int bio_index,
                          sector_t sector,
                          int sectorcnt,
                          char *buffer,
                          int is_write)
                  {
#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA
#warning Using interrupt based SDMA
                          int ret;

                          if (is_write)
                          {
                                  /*
                                   * for TX beter wait the drive interrupt
                                   * because we first write into drive buffer and the drive do the real stuff
                                  */
                                  sdma_ata_tx_init(priv->sdma);

                                  ret = start_writerequest(priv, req, bio, bio_index, sector, sectorcnt, (u16 *) buffer);
                          } else
                          {
                                  /*
                                   * for RX better write for sdma int
                                   * because the drive will first to the read and then the SDMA task fetch
                                   * the data from the FIFO
                                  */
                                  sdma_ata_rx_init(priv->sdma);

                                  // Wait for int before starting the task
                                  ret = start_readrequest(priv, req, bio, bio_index, sector, sectorcnt, (u16 *) buffer);
                          }

                          return ret;

#else /* CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA */
#warning Using poll based SDMA -> Slow and obsolet!
                          return
                                  is_write
                                  ? do_writerequest(priv, req, bio, bio_index, sector, sectorcnt, (u16 *) buffer)
                                  : do_readrequest(priv, req, bio, bio_index, sector, sectorcnt, (u16 *) buffer);
#endif /* CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA */
                  }



                  /**************/
                  /**************/
                  int mpc52xx_ata_docpupollread(
                          struct mpc52xx_blockata_priv *priv,
                          void *buffer,
                          int len)
                  {
                          u16 *buffer16 = (u16 *) buffer;
                          int local_len = len;

                          while(local_len--)
                                  *buffer16++ = read_data(priv);

                          return len - local_len;
                  }

#ifdef CONFIG_BLK_DEV_MPC52XX_ATAPIO_SDMA
#if 0
                  static int ata_dosdmapollwrite(
                          struct mpc52xx_blockata_priv *priv,
                          void *buffer,
                          int len)
                  {
                          int ret = 0;
                          u32 port_BusAddr;
                          u32 addr_BusAddr;
                          int taskirq;
                          int timeout = 10*1000;

                  //NPRINTK("%s: have to read %d , from port 0x%8.8lx to mem %p\n", __func__, count, port, addr);

                          taskirq = priv->sdma_irq;
                          port_BusAddr = ATAFIFO_BUSADDR;

                          addr_BusAddr = virt_to_phys( (void*) buffer);

                          //NPRINTK("%s:priv=%p, irq=%d, port bus=0x%8.8x, addr bus =0x%8.8x\n",
                          //  __func__, priv, taskirq, port_BusAddr, addr_BusAddr)

                          // Turn the SDMA into RX
                          sdma_ata_tx_init(priv->sdma);

                          sdma_ata_submit_buffer(priv->sdma, (void *)buffer, (void *) addr_BusAddr, (void *)port_BusAddr, len*2);

                          sdma_ata_clear_irq(priv->sdma);
                          sdma_ata_enable(priv->sdma);

                          for(;;)
                          {

                                  u32 val;
                                  val = in_be32(&sdma.io->IntPend);
                                  if ( (val & (1 << priv->sdma->tasknum) ) )
                                          break;

                                  if (timeout--<=0) {
                                          ret = -1;
                                          printk("timeout 0x%x\n", read_altstatus(priv) );
                                          break;
                                  }

                                  udelay(1);
                          }

                          sdma_ata_disable(priv->sdma);
                          sdma_ata_reset(priv->sdma);

                          return 0;
                  }

                  static int ata_dosdmapollread(
                          struct mpc52xx_blockata_priv *priv,
                          void *buffer,
                          int len)
                  {
                          int ret = 0;
                          u32 port_BusAddr;
                          u32 addr_BusAddr;
                          int taskirq;
                          int timeout = 10*1000;

                          taskirq = priv->sdma_irq;
                          port_BusAddr = ATAFIFO_BUSADDR;
                          addr_BusAddr = virt_to_phys( (void*) buffer);

                          // Turn the SDMA into RX
                          sdma_ata_rx_init(priv->sdma);

                          sdma_ata_submit_buffer(priv->sdma, (void *)buffer, (void *)port_BusAddr, (void *) addr_BusAddr, len*2);

                          sdma_ata_clear_irq(priv->sdma);
                          sdma_ata_enable(priv->sdma);

                          for(;;)
                          {
                                  u32 val;
                                  val = in_be32(&sdma.io->IntPend);
                                  if ( (val & (1 << priv->sdma->tasknum) ) )
                                          break;

                                  if (timeout--<=0) {
                                          printk("timeout 0x%x\n", read_altstatus(priv) );
                                          ret = -1;
                                          break;
                                  }

                                  udelay(1);
                          }

                          sdma_ata_disable(priv->sdma);
                          sdma_ata_reset(priv->sdma);

                          return ret;
                  }
#endif
#endif

                  /**************/
                  int mpc52xx_ata_docpupollwrite(
                          struct mpc52xx_blockata_priv *priv,
                          void *buffer,
                          int len)
                  {
                          u16 *buffer16 = (u16 *) buffer;
                          int local_len = len;

                          while(local_len--)
                                  write_data(priv, *buffer16++);

                          return len - local_len;
                  }

/*
 *
*/

EXPORT_SYMBOL(mpc52xx_ata_docpupollread);
EXPORT_SYMBOL(mpc52xx_ata_docpupollwrite);
EXPORT_SYMBOL(mpc52xx_ata_dotransfer);
EXPORT_SYMBOL(sdma_void_handler);
EXPORT_SYMBOL(ata_void_handler);
EXPORT_SYMBOL(mpc52xx_ata_ack_blkreq);
